@@ -241,6 +241,8 @@ use minus_core::events::Event;
 use minus_core::search;
 #[cfg(feature = "search")]
 pub use minus_core::search::SearchMode;
+#[cfg(feature = "search")]
+use std::collections::BTreeSet;
 use std::string::ToString;
 use std::{fmt, io::stdout};
 
@@ -510,6 +512,10 @@ pub struct PagerState {
     formatted_lines: Vec<String>,
     /// Configuration for line numbers. See [`LineNumbers`]
     pub line_numbers: LineNumbers,
+    /// Unterminated lines
+    /// Keeps track of the number of lines at the last of [PagerState::formatted_lines] which are
+    /// not terminated by a newline
+    unterminated: usize,
     /// The prompt displayed at the bottom wrapped to available terminal width
     prompt: Vec<String>,
     /// The input classifier to be called when a input is detected
@@ -546,8 +552,9 @@ pub struct PagerState {
     #[cfg_attr(docsrs, cfg(feature = "search"))]
     pub search_mode: SearchMode,
     /// Lines where searches have a match
+    /// In order to avoid duplicate entries of lines, we keep it in a [`BTreeSet`]
     #[cfg(feature = "search")]
-    search_idx: Vec<usize>,
+    search_idx: BTreeSet<usize>,
     /// Index of search item currently in focus
     /// It should be 0 even when no search is in action
     #[cfg(feature = "search")]
@@ -582,14 +589,15 @@ impl PagerState {
         };
 
         Ok(Self {
-            lines: String::new(),
-            formatted_lines: Vec::new(),
+            lines: String::with_capacity(u16::MAX.into()),
+            formatted_lines: Vec::with_capacity(u16::MAX.into()),
             line_numbers: LineNumbers::Disabled,
             upper_mark: 0,
+            unterminated: 0,
             prompt: wrap_str("minus", cols),
             exit_strategy: ExitStrategy::ProcessQuit,
             input_classifier: Box::new(input::DefaultInputClassifier {}),
-            exit_callbacks: Vec::new(),
+            exit_callbacks: Vec::with_capacity(5),
             message: None,
             #[cfg(feature = "static_output")]
             run_no_overflow: false,
@@ -598,7 +606,7 @@ impl PagerState {
             #[cfg(feature = "search")]
             search_mode: SearchMode::default(),
             #[cfg(feature = "search")]
-            search_idx: Vec::new(),
+            search_idx: BTreeSet::new(),
             #[cfg(feature = "search")]
             search_mark: 0,
             // Just to be safe in tests, keep at 1x1 size
@@ -612,22 +620,48 @@ impl PagerState {
         self.formatted_lines.len()
     }
 
+    /// Formats the given `line`
+    ///
+    /// - `line_numbers` tells whether to format the line with line numbers.
+    /// - `len_line_number` is the length of the number of lines in [`PagerState::lines`] as in a string.
+    ///     For example, this will be 2 if number of lines in [`PagerState::lines`] is 50 and 3 if
+    ///     number of lines in [`PagerState::lines`] is 500. This is used for calculating the padding
+    ///     of each displayed line.
+    /// - `idx` is the position index where the line is placed in [`PagerState::lines`].
+    /// - `formatted_idx` is the position index where the line will be placed in the resulting
+    ///    [`PagerState::formatted_lines`]
     pub(crate) fn formatted_line(
         &self,
         line: &str,
         line_numbers: bool,
         len_line_number: usize,
         idx: usize,
+        #[cfg(feature = "search")] formatted_idx: usize,
+        #[cfg(feature = "search")] search_idx: &mut BTreeSet<usize>,
     ) -> Vec<String> {
         if line_numbers {
+            // Padding is the space that the actual line text will be shifted to accomodate for
+            // in line numbers. This is equal to:-
+            // 1 for initial space + len_line_number + 1 for `.` sign and + 1 for the followup space
+            //
+            // We reduce this from the number of available columns as this space cannot be used for
+            // actual line display when wrapping the lines
+            let padding = len_line_number + 3;
             #[cfg_attr(not(feature = "search"), allow(unused_mut))]
-            wrap_str(line, self.cols.saturating_sub(len_line_number + 3))
+            #[cfg_attr(not(feature = "search"), allow(unused_variables))]
+            wrap_str(line, self.cols.saturating_sub(padding))
                 .into_iter()
-                .map(|mut row| {
+                .enumerate()
+                .map(|(wrap_idx, mut row)| {
                     #[cfg(feature = "search")]
                     if let Some(st) = self.search_term.as_ref() {
                         // highlight the lines with matching search terms
-                        row = search::highlight_line_matches(&row, st);
+                        // If a match is found, add this line's index to PagerState::search_idx
+                        let (hrow, is_match) = search::highlight_line_matches(&row, st);
+                        if is_match {
+                            search_idx.insert(formatted_idx + wrap_idx);
+                        }
+                        row = hrow;
                     }
 
                     if cfg!(not(test)) {
@@ -640,6 +674,8 @@ impl PagerState {
                             row = row
                         )
                     } else {
+                        // In tests, we don't care about ANSI sequences for cool looking line numbers
+                        // hence we don't include them in tests. It just makes testing more difficult
                         format!(
                             " {number: >len$}. {row}",
                             number = idx + 1,
@@ -650,26 +686,53 @@ impl PagerState {
                 })
                 .collect::<Vec<String>>()
         } else {
-            #[cfg(feature = "search")]
-            if let Some(st) = self.search_term.as_ref() {
-                return wrap_str(&search::highlight_line_matches(line, st), self.cols);
-            }
-
+            #[cfg_attr(not(feature = "search"), allow(unused_variables))]
             wrap_str(line, self.cols)
+                .iter()
+                .enumerate()
+                .map(|(wrap_idx, row)| {
+                    #[cfg(feature = "search")]
+                    {
+                        self.search_term.as_ref().map_or_else(
+                            || row.to_string(),
+                            |st| {
+                                // highlight the lines with matching search terms
+                                // If a match is found, add this line's index to PagerState::search_idx
+                                let (hrow, is_match) = search::highlight_line_matches(row, st);
+                                if is_match {
+                                    search_idx.insert(formatted_idx + wrap_idx);
+                                }
+                                hrow
+                            },
+                        )
+                    }
+                    #[cfg(not(feature = "search"))]
+                    row.to_string()
+                })
+                .collect::<Vec<String>>()
         }
     }
 
     pub(crate) fn format_lines(&mut self) {
+        // Keep it for the record and don't call it unless it is really necessory as this is kinda
+        // expensive
         let line_count = self.lines.lines().count();
 
+        // Calculate len_line_number. This will be 2 if line_count if 50 and 3 if line_count is 100.
         let len_line_number = line_count.to_string().len();
 
+        // Search idx, this will get filled by the self.formatted_line function
+        // we will later set this to self.search_idx
+        #[cfg(feature = "search")]
+        let mut search_idx = BTreeSet::new();
+        let mut formatted_idx = 0;
+
         self.formatted_lines = self
             .lines
             .lines()
             .enumerate()
             .flat_map(|(idx, line)| {
-                self.formatted_line(
+                let new_line = self.formatted_line(
                     line,
                     matches!(
                         self.line_numbers,
@@ -677,17 +740,26 @@ impl PagerState {
                     ),
                     len_line_number,
                     idx,
-                )
+                    #[cfg(feature = "search")]
+                    formatted_idx,
+                    #[cfg(feature = "search")]
+                    &mut search_idx,
+                );
+                formatted_idx += new_line.len();
+                new_line
             })
             .collect::<Vec<String>>();
 
+        #[cfg(feature = "search")]
+        {
+            self.search_idx = search_idx;
+        }
+
+        // Wrap any message if present and also the prompt
         if self.message.is_some() {
             rewrap(self.message.as_mut().unwrap(), self.cols);
         }
         rewrap(&mut self.prompt, self.cols);
-
-        #[cfg(feature = "search")]
-        search::set_match_indices(self);
     }
 
     /// Returns all the text within the bounds, after flattening
@@ -709,53 +781,54 @@ impl PagerState {
     }
 
     pub(crate) fn append_str(&mut self, text: &str) {
-        let mut fmt_line = self.make_append_str(text);
-        self.formatted_lines.append(&mut fmt_line);
+        let (fmt_line, num_unterminated) = self.make_append_str(text);
+        self.append_str_on_unterminated(fmt_line, num_unterminated);
     }
 
-    pub(crate) fn make_append_str(&mut self, text: &str) -> Vec<String> {
-        // if the text we have saved currently ends with a newline,
+    /// Makes the text that will be displayed and appended it to [`self.formatted_lines`]
+    ///
+    /// - The first output value is the actual text rows that needs to be appended. This is wrapped
+    ///     based on the available columns
+    /// - The second value is the number of rows that should be truncated from [`self.formatted_lines`]
+    ///     before appending this line. This will be 0 if the given `text` is to be appended to
+    ///     [`self.formatted_lines`] but will be `>0` if the given text is actually part of the
+    ///     last appended line. This function determines this by checking whether self.lines ends with
+    ///     `\n` after appending the text
+    pub(crate) fn make_append_str(&mut self, text: &str) -> (Vec<String>, usize) {
+        // if the text we have saved currently in self.lines ends with a newline or is empty,
         // we want the formatted_text vector to append the line instead of
-        // trying to add it to the last item
-        let newline = self.lines.ends_with('\n');
-
-        // find the number of trailing whitespace characters currently on self.lines
-        let ending_whitespace = self
-            .lines
-            .chars()
-            .rev()
-            .take_while(|c| c.is_whitespace() && *c != '\n')
-            .collect::<String>();
+        // trying to add it to the last item.
+        //
+        // In minus the \n acts as a marker that a line has been terminated and no changes are going
+        // to be made to it again. It may or may not be present at the end of the last line of self.lines
+        // If it is, we know that no changes are going to be made to it. In case it isn't, minus believes
+        // that the incoming text is part of the last line and hence here need to check that.
+        let append = self.lines.ends_with('\n') || self.lines.is_empty();
 
         // push the text to lines
         self.lines.push_str(text);
 
+        let to_skip = self.lines.lines().count();
+
         // And get how many lines of text will be shown (not how many rows, how many wrapped
         // lines), and get its string length
-        let len_line_number = self.lines.lines().count().to_string().len();
+        let len_line_number = to_skip.to_string().len();
 
-        // if we want a newline, just format the new text and append it.
-        // if we don't, format the text with the last line currently formatted
-        // since it will be appended to that
-        //
-        // also get the line number to start at when formatting
-        let (to_format, to_skip) = if newline {
-            (text.to_owned(), self.lines.lines().count())
+        // If append is true, we take the text for formatting
+        // else we take the last line of self.lines for formatting. This is because we nned to
+        // format the entire line rathar than just this part
+        let to_format = if append {
+            text.to_owned()
         } else {
-            // add the trailing whitespace in here, since it isn't preserved when wrapping the
-            // lines, and thus won't appear on the last element in self.formatted_lines
-            let to_fmt = format!(
-                "{}{}{}",
-                self.formatted_lines.pop().unwrap_or_default(),
-                ending_whitespace,
-                text
-            );
-
-            (to_fmt, self.lines.lines().count().saturating_sub(1))
+            self.lines.lines().last().unwrap_or_default().to_string()
         };
 
+        // This will get filled if there is an ongoing search. We just need to append it to
+        // self.search_idx at the end
+        #[cfg(feature = "search")]
+        let mut append_search_idx = BTreeSet::new();
         // format the lines we want to format
-        to_format
+        let formatted_text = to_format
             .lines()
             .enumerate()
             .flat_map(|(idx, line)| {
@@ -767,9 +840,47 @@ impl PagerState {
                     ),
                     len_line_number,
                     idx + to_skip.saturating_sub(1),
+                    #[cfg(feature = "search")]
+                    if append {
+                        self.formatted_lines.len()
+                    } else {
+                        self.formatted_lines.len().saturating_sub(1)
+                    },
+                    #[cfg(feature = "search")]
+                    &mut append_search_idx,
                 )
             })
-            .collect::<Vec<String>>()
+            .collect::<Vec<String>>();
+        #[cfg(feature = "search")]
+        self.search_idx.append(&mut append_search_idx);
+        let fmt_text_len = formatted_text.len();
+
+        (
+            formatted_text,
+            if self.lines.ends_with('\n') {
+                0
+            } else {
+                fmt_text_len
+            },
+        )
+    }
+
+    /// Conditionally appends to [`self.formatted_lines`] or changes the last unterminated rows of
+    /// [`self.formatted_lines`]
+    ///
+    /// `num_unterminated` is the current number of lines returned by [`self.make_append_str`]
+    /// that should be truncated from [`self.formatted_lines`] to update the last line
+    pub(crate) fn append_str_on_unterminated(
+        &mut self,
+        mut fmt_line: Vec<String>,
+        num_unterminated: usize,
+    ) {
+        if num_unterminated != 0 || self.unterminated != 0 {
+            self.formatted_lines
+                .truncate(self.formatted_lines.len() - self.unterminated);
+        }
+        self.formatted_lines.append(&mut fmt_line);
+        self.unterminated = num_unterminated;
     }
 }
 
